from typer.testing import CliRunner

from kurra.cli import app
from kurra.cli.console import console

runner = CliRunner()

# g = Graph().parse(
#     data="""
#     [
#       {
#         "@context": {
#           "olis": "https://olis.dev/",
#           "RealGraph": "olis:RealGraph",
#           "VirtualGraph": "olis:VirtualGraph",
#           "hasGraphRole": "olis:hasGraphRole",
#           "isAliasFor": "olis:isAliasFor",
#           "gr": "https://olis.dev/GraphRoles/"
#         },
#         "@graph": {
#           "@id": "https://example.org/library",
#           "@type": "RealGraph",
#           "hasGraphRole": {"@id": "gr:Original"}
#         }
#       }
#     ]
#     """,
#     format="json-ld",
# )
#
# g.bind("olis", "https://olis.dev/")
# g.bind("gr", "https://olis.dev/GraphRoles/")
#
# print(g.serialize(format="longturtle"))

result = runner.invoke(
    app,
    [
        "shacl",
        "syncv",
    ],
)
# print(result.output)
#
# print("xxxxxx")

# result = runner.invoke(
#     app,
#     [
#         "shacl",
#         "listv",
#     ],
# )


# from kurra.shacl import list_local_validators
#
# print(list_local_validators())

# print(sync_validators())

# result = runner.invoke(
#     app,
#     [
#         "shacl",
#         "validate",
#         str(Path(__file__).parent / "tests/test_shacl/vocab-invalid.ttl"),
#         "8"
#     ],
# )

if result.exception:
    console.print(result.exception)
else:
    console.print(result.output)

# print(validate(Path(__file__).parent / "tests/test_shacl/vocab-valid.ttl", 8))



# OLIS API

from rdflib import URIRef
from rdflib.namespace import DefinedNamespace, Namespace

OLIS = Namespace("https://olis.dev/")


class OLIS_GRAPH_ROLES(DefinedNamespace):
    _NS = Namespace("http://olis.dev/GraphRoles/")
    _fail = True

    Original: URIRef
    Inferred: URIRef
    Added: URIRef
    Removed: URIRef



from pathlib import Path
from pickle import dump, load

import httpx
from pyshacl import validate as v
from rdflib import Dataset, Graph, URIRef
from rdflib.namespace import SDO

from kurra.db.gsp import get as gsp_get
from kurra.sparql import query
from kurra.utils import load_graph


def validate(
    data_file_or_dir_or_graph: Path | Graph,
    shacl_graph_or_file_or_url_or_id: Graph | Path | str | int,
) -> tuple[bool, Graph, str]:
    """Runs pySHACL's validate() function with some preset values"""
    data_graph = load_graph(data_file_or_dir_or_graph)
    shapes_graph = get_validator_graph(shacl_graph_or_file_or_url_or_id)
    if shapes_graph is None:
        raise RuntimeError(
            f"Not able to load shapes graph: {shacl_graph_or_file_or_url_or_id}"
        )

    return v(data_graph, shacl_graph=shapes_graph, allow_warnings=True)


def list_local_validators() -> dict[str, dict[str, int]] | None:
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"
    validator_ids_cache = kurra_cache / "validator_ids.pkl"

    if Path.is_file(validators_cache):
        local_validators = {}
        cv = load(open(validators_cache, "rb"))
        cv: Dataset
        validator_iris = [
            x.identifier
            for x in cv.graphs()
            if str(x.identifier) not in ["urn:x-rdflib:default"]
        ]

        validator_ids = load(open(validator_ids_cache, "rb"))

        for validator_iri in sorted(validator_iris):
            validator_id = validator_ids[validator_iri]
            validator_name = load_graph(cv.get_graph(validator_iri)).value(
                subject=validator_iri, predicate=SDO.name
            )
            local_validators[str(validator_iri)] = {
                "name": str(validator_name),
                "id": str(validator_id),
            }

        return local_validators
    else:
        return {}


def sync_validators(http_client: httpx.Client | None = None):
    """Checks the Semantic Background, currently https://fuseki.dev.kurrawong.ai/semback, for known validators.

    It then checks local storage to see which, if any, of those validators are stored locally.

    For any missing, it pulls down and stores a copy locally and updates the known list of available validators.
    """
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"
    validator_ids_cache = kurra_cache / "validator_ids.pkl"
    semback_sparql_endpoint = "https://api.data.kurrawong.ai/sparql"

    # get list of remote validators
    q = """
        PREFIX schema: <https://schema.org/>
        
        SELECT * 
        WHERE { 
          <https://data.kurrawong.ai/sb/validators> schema:hasPart ?p
        }
        """
    r = query(semback_sparql_endpoint, q, None, http_client, "python", True)

    remote_validators = [row["p"] for row in r]

    # get list of local validators
    local_validators = list_local_validators()

    # diff the lists
    unknown_validators = list(set(remote_validators) - set(local_validators.keys()))

    # prepare to cache
    if len(unknown_validators) > 0:
        if not kurra_cache.exists():
            Path(kurra_cache).mkdir()

        # get & add unknown remote validators to local
        if validators_cache.exists():
            d = load(open(validators_cache, "rb"))
        else:
            d = Dataset()

        for v in unknown_validators:
            g = gsp_get(semback_sparql_endpoint, v, http_client=http_client)
            if g == 422:
                raise NotImplementedError(
                    "The KurrawongAI Semantic Background set of validators is not available yet."
                )
            if not isinstance(g, Graph):
                raise RuntimeError(
                    f"The graph {v} was not obtained from the SPARQL Endpoint {semback_sparql_endpoint}"
                )
            d.add_graph(g)
            print(f"Caching validator {g.identifier}")

        with open(validators_cache, "wb") as f:
            dump(d, f)

        validator_ids = {}
        for i, v in enumerate(sorted([x.identifier for x in d.graphs()])):
            validator_ids[v] = i + 1

        with open(validator_ids_cache, "wb") as f2:
            print("Dumping validator IDs")
            dump(validator_ids, f2)

    local_validators = list_local_validators()

    return local_validators


def get_validator_graph(
    graph_or_file_or_url_or_id: Graph | Path | str | int,
) -> Graph | None:
    kurra_cache = Path().home() / ".kurra"
    validators_cache = kurra_cache / "validators.pkl"
    validator_ids_cache = kurra_cache / "validator_ids.pkl"

    # it's a local ID so look it up in cache
    if isinstance(graph_or_file_or_url_or_id, int) or (
        isinstance(graph_or_file_or_url_or_id, str)
        and graph_or_file_or_url_or_id.isdigit()
    ):
        validator_ids = load(open(validator_ids_cache, "rb"))
        validator_iris = [
            key
            for key, value in validator_ids.items()
            if value == int(graph_or_file_or_url_or_id)
        ]
        if len(validator_iris) != 1:
            raise ValueError(
                f"Could not find validator for {graph_or_file_or_url_or_id}"
            )

        cv = load(open(validators_cache, "rb"))
        cv: Dataset
        return cv.graph(URIRef(validator_iris[0]))

    # cater for CLI making paths strings
    if isinstance(graph_or_file_or_url_or_id, str):
        if Path(graph_or_file_or_url_or_id).exists():
            return load_graph(Path(graph_or_file_or_url_or_id))

    try:
        return load_graph(graph_or_file_or_url_or_id)
    except:
        return None



from pathlib import Path

import typer
from rich.table import Table

from kurra.cli.console import console
from kurra.cli.utils import (
    format_shacl_graph_as_rich_table,
)
from kurra.shacl import list_local_validators, sync_validators, validate

app = typer.Typer(help="SHACL commands")


@app.command(
    name="validate",
    help="Validate a given file or directory of RDF files using a given SHACL file or directory of files",
)
def validate_command(
    file_or_dir: Path = typer.Argument(
        ..., help="The file or directory of RDF files to be validated"
    ),
    shacl_graph_or_file_or_url_or_id: str = typer.Argument(
        ..., help="The file or directory of SAHCL files to validate with"
    ),
) -> None:
    """Validate a given file or directory of files using a given SHACL file or directory of files"""
    valid, g, txt = validate(file_or_dir, shacl_graph_or_file_or_url_or_id)
    if valid:
        console.print("The data is valid")
    else:
        console.print("The data is NOT valid")
        console.print("The errors are:")
        console.print(format_shacl_graph_as_rich_table(g))


@app.command(
    name="listv",
    help="Lists all known SHACL validators",
)
def listv_command():
    l = list_local_validators()
    if l is None:
        console.print("No local validators found")
    else:
        t = Table()
        t.add_column("ID")
        t.add_column("IRI")
        t.add_column("Name")
        for k, v in list_local_validators().items():
            t.add_row(v["id"], k, v["name"])
        console.print(t)


@app.command(
    name="syncv",
    help="Synchronizes SHACL validators",
)
def syncv_command():
    sync_validators()

    console.print("Synchronizing SHACL validators")



import json
from pathlib import Path
from typing import Literal

import httpx
from rdflib import Dataset, Graph

from kurra.db.sparql import query as db_query
from kurra.utils import (
    add_namespaces_to_query_or_data,
    convert_sparql_json_to_python,
    load_graph,
    make_sparql_dataframe,
)


def query(
    p: Path | str | Graph | Dataset,
    q: str,
    namespaces: dict[str, str] | None = None,
    http_client: httpx.Client = None,
    return_format: Literal["original", "python", "dataframe"] = "original",
    return_bindings_only: bool = False,
):
    """Pose a SPARQL query to a file, and RDF Graph or a SPARQL Endpoint"""
    if p is None:
        raise ValueError(
            "You must supply a Path, string (of data or a URL), Graph or a Dataset to query for variable p"
        )

    if q is None:
        raise ValueError("You must supply a query")

    if return_format not in ["original", "python", "dataframe"]:
        raise ValueError(
            f"return_format {return_format} must be either 'original', 'python' or 'dataframe'"
        )

    if namespaces is not None:
        q = add_namespaces_to_query_or_data(q, namespaces)

    if http_client is None:
        http_client = httpx.Client()

    if return_format == "dataframe":
        if (
            "CONSTRUCT" in q
            or "DESCRIBE" in q
            or "INSERT" in q
            or "DELETE" in q
            or "DROP" in q
        ):
            raise ValueError(
                'Only SELECT and ASK queries can have return_format set to "dataframe"'
            )

        try:
            from pandas import DataFrame
        except ImportError:
            raise ValueError(
                'You selected the output format "dataframe" by the pandas Python package is not installed.'
            )

    if "CONSTRUCT" in q or "DESCRIBE" in q:
        s = None
        f = None
        if isinstance(p, str) and p.startswith("http"):
            r = db_query(p, q, namespaces, http_client, "original", False)
            s = load_graph(r)

        else:  # (isinstance(p, str) and not p.startswith("http")) or isinstance(p, Path):
            f = load_graph(p).query(q)

        if return_format == "dataframe":
            raise ValueError(
                "DataFrames cannot be returned for CONSTRUCT or DESCRIBE queries"
            )
        elif return_format == "python":
            return s if s is not None else f.graph
        else:
            return (
                s.serialize(format="longturtle")
                if s is not None
                else f.graph.serialize(format="longturtle")
            )

    elif "INSERT" in q or "DELETE" in q:
        raise NotImplementedError(
            "INSERT & DELETE queries are not yet implemented by this interface. Try kurra.db.sparql"
        )

    elif "DROP" in q:
        if isinstance(p, str) and p.startswith("http"):
            r = db_query(p, q, namespaces, http_client, return_format, False)

            if r == "":
                return ""
        else:
            raise NotImplementedError("DROP commands are not yet implemented for files")

    else:  # SELECT or ASK
        close_http_client = False
        if http_client is None:
            http_client = httpx.Client()
            close_http_client = True

        r = None
        if isinstance(p, str) and p.startswith("http"):
            r = db_query(
                p, q, namespaces, http_client, return_format, return_bindings_only
            )

        if close_http_client:
            http_client.close()

        if r is not None:
            return r
        else:  # querying a file or string RDF data
            r = load_graph(p).query(q).serialize(format="json")

            if return_format == "dataframe":
                return make_sparql_dataframe(json.loads(r))
            elif return_format == "python":
                return convert_sparql_json_to_python(r, return_bindings_only)
            else:
                return r.decode()
